# chat/consumers.py
import json

from channels.auth import login
from channels.db import database_sync_to_async
from channels.generic.websocket import AsyncWebsocketConsumer
from django import dispatch
from rest_framework.authentication import TokenAuthentication

message_received = dispatch.Signal()


class EventConsumer(AsyncWebsocketConsumer):
    async def connect(self):
        self.user = None
        await self.accept()
        await self.join_groups()

    async def disconnect(self, close_code):
        await self.leave_groups()

    async def receive(self, text_data):
        [event, msg] = json.loads(text_data)

        # heartbeat
        if event == 'heartbeat':
            await self.send(text_data=text_data)
            await self.join_groups()

        # echo
        elif event == 'echo':
            # await
            user = self.scope.get('user')
            if user:
                await self.channel_layer.group_send(
                    f'user.{user.id}',
                    dict(type='distribute', message=text_data)
                )
            else:
                await self.send(text_data=text_data)

        # login
        elif event == 'login':
            await self.login(msg)

        # user
        elif event == 'user':
            user = self.scope.get('user')
            await self.send(text_data=json.dumps([event, dict(
                msgid=msg.get('msgid'),
                data=user and user.id or None,
            )]))

        # everything else
        else:
            await self.send_other(event, msg)

    @database_sync_to_async
    def send_other(self, event: str, msg: dict):
        message_received.send(
            sender=self,
            user=self.scope['user'],
            event=event,
            msg=msg,
        )

    async def distribute(self, packet: dict):
        await self.send(text_data=packet['message'])

    @database_sync_to_async
    def validate_user_token(self, token: str):
        user = None
        try:
            user, token = TokenAuthentication().authenticate_credentials(token)
        except Exception:
            pass

        return user

    async def join_groups(self):
        await self.channel_layer.group_add('everyone', self.channel_name)

        user = self.scope.get('user')
        if user:
            await self.channel_layer.group_add(f'user.{user.id}', self.channel_name)
            async for group in user.groups.all():
                await self.channel_layer.group_add(f'group.{group.id}', self.channel_name)

    async def leave_groups(self):
        await self.channel_layer.group_discard('everyone', self.channel_name)

        user = self.scope.get('user')
        if user:
            await self.channel_layer.group_discard(f'user.{user.id}', self.channel_name)
            async for group in user.groups.all():
                await self.channel_layer.group_discard(f'group.{group.id}', self.channel_name)

    async def login(self, msg: dict):
        token = msg.get('token')
        if token:
            user = await self.validate_user_token(token)

            if user:
                user.backend = 'virgodev_websocket_support'
                await login(self.scope, user)
                await database_sync_to_async(self.scope['session'].save)()
                self.scope['user'] = user
                await self.join_groups()

                await self.send(json.dumps([
                    'login', dict(ok=True, msgid=msg.get('msgid'))
                ]))
                return

        await self.send(json.dumps([
            'login', dict(error="Token missing or invalid")
        ]))
